
----------------------------------------
-- Variables
----------------------------------------

local _addonName, _addonData = ...;
local _L = _addonData.L;
local _version = GetAddOnMetadata(_addonName, "version");
local _aVar = _addonData.variables;

local FORMAT_OFFSPEC = "|cFFAA0000%s|r"
local FORMAT_OFFSPEC_ICON =  CreateAtlasMarkup("Map-MarkedDefeated", 12, 12, 0, 0)

local ILWAddon = LibStub("AceAddon-3.0"):NewAddon(_addonName);
local ILWBroker = LibStub("LibDataBroker-1.1"):NewDataObject(_addonName, {
	type = "launcher",
	text = "I Learned What?",
	icon = "Interface/ICONS/Spell_Holy_SurgeOfLight",
	OnClick = function(self, button, down)
		if (ILW_UnlockContainer:IsShown()) then
			HideUIPanel(ILW_UnlockContainer);
		else
			ILW_UnlockContainer:Open();
		end
	end,
	OnTooltipShow = function(tt)
		tt:AddLine("I Learned What?", 1, 1, 1);
		tt:AddLine(_L["MINIMAP_TOOLTIP"]);
	end	
})
local _icon = LibStub("LibDBIcon-1.0");
local ADD = LibStub("AddonDropDown-1.0");

local _defaults = {
	["global"] = {	
		showPopup = true,
		minimap = {
			hide = false,
		},
		["trackers"] = {
			["Spells"] = true
			,["Talents"] = true
			,["Dungeons"] = true
			,["Riding"] = true
			,["UI"] = true
			,["Battlegrounds"] = true
			,["Zones"] = true
			,["Misc"] = true
			,["OffSpec"] = false
		}
		,["output"] = {};
	}
	,["char"] = {
		["specLastLevel"] = {},
	}
}

local _trackers = {
		["Spells"] = true
		,["Talents"] = true
		,["Dungeons"] = true
		,["Riding"] = true
		,["Battlegrounds"] = true
		,["UI"] = true
		,["Zones"] = true
		,["Misc"] = true
		,["OffSpec"] = false
	}
	
local _trackerLabels = {
		["Spells"] = SPELLS
		,["Talents"] = BONUS_TALENTS
		,["Dungeons"] = _L["TRACKER_INSTANCES"]
		,["Riding"] = _L["TRACKER_RIDING"]
		,["Battlegrounds"] = BATTLEGROUNDS
		,["UI"] = UIOPTIONS_MENU
		,["Zones"] = _L["TRACKER_ZONES"]
		,["Misc"] = BINDING_HEADER_COMMENTATORMISC
		,["OffSpec"] = _L["TRACKER_OFFSPEC"];
	}

_addonData.help = {}
local _help = _addonData.help;


local UNLOCKTYPE_FILLER = -1;
local UNLOCKTYPE_SPELL = 1;
local UNLOCKTYPE_TALENT = 2;
local UNLOCKTYPE_UI = 3;
local UNLOCKTYPE_RIDING = 4;
local UNLOCKTYPE_MISC = 5;
local UNLOCKTYPE_ZONES = 6;
local UNLOCKTYPE_DUNGEON = 7;
local UNLOCKTYPE_RAID = 8;
local UNLOCKTYPE_PVP = 9;
local UNLOCKTYPE_TUTORIAL = 10;
local UNLOCKTYPE_UPCOMMING = 99;

local UNLOCK_SUBTYPE_TUTORIAL = 0;
local UNLOCK_SUBTYPE_UPCOMMING = 1;

local UNLOCK_UPCOMMING_TRUNC = {["name"] = "+x", ["type"] = UNLOCKTYPE_UPCOMMING, ["subType"] = UNLOCK_SUBTYPE_UPCOMMING, ["noTooltip"] = true, ["id"] = 0, ["icon"] = "Interface/ICONS/Spell_ChargePositive", ["extra"] = {}, validSpec="true"};

local ICON_TALENT = "Interface/ICONS/Ability_Marksmanship";
local ICON_UI = "Interface/ICONS/INV_Scroll_02";

local UNLOCKS_PER_PAGE = 10;

local MAXLEVEL = GetMaxPlayerLevel();
local MINLEVEL = 1;
local _playerClass = select(2, UnitClass("player"));
if (_playerClass == "DEATHKNIGHT") then
	MINLEVEL = 55
elseif (_playerClass == "DEMONHUNTER") then
	MINLEVEL = 98
end

local _helpPlate = {
	FramePos = { x = 5,	y = -25 },
	FrameSize = { width = 440, height = 495	},
	[1] = { ButtonPos = { x = 325,	y = -115}, HighLightBox = { x = 43, y = -60, width = 570, height = 375 }, ToolTipDir = "DOWN", ToolTipText = _L["HELP_INFO"] }
	,[2] = { ButtonPos = { x = 321,	y = -445}, HighLightBox = { x = 10, y = -445, width = 620, height = 70 }, ToolTipDir = "UP", ToolTipText = _L["HELP_INFO_CHART"] }
}
local _typeIcons = {
		[UNLOCKTYPE_SPELL] = "Class"
		,[UNLOCKTYPE_UI] = "poi-workorders"
		,[UNLOCKTYPE_DUNGEON] = "Dungeon"
		,[UNLOCKTYPE_RAID] = "Raid"
		,[UNLOCKTYPE_TALENT] = "Profession"
		,[UNLOCKTYPE_PVP] = "CrossedFlags"
		,[UNLOCKTYPE_RIDING] = "StableMaster"
		,[UNLOCKTYPE_ZONES] = "QuestNormal"
	};

function ILW_ShowHelpUnlocks(show)	
	ILW_UnlockContainer:UpdateUnlockDisplay();
end

local function GetPlayerSpec()
	if (GetSpecialization() ~= nil) then
		return GetSpecializationInfo(GetSpecialization());
	end
	
	return nil;
end

local function IsValidSpec(unlock)
	if (unlock.type ~= UNLOCKTYPE_SPELL or not unlock.specs) then return true; end
	local specId =  GetPlayerSpec();
	for k, spec in ipairs(unlock.specs) do
		if (spec == specId) then
			return true;
		end
	end
	return false;
end

local function AddUnlocksToTooltip(tooltip, unlocks)
	local lastType;
	local iconString = "";

	for k, unlock in ipairs(unlocks) do
		if (unlock.type ~= lastType) then
		
			local texSize = 512;
			local atlas = unlock.atlas or _typeIcons[unlock.type];
			if atlas then
				local atlasInfo = C_Texture.GetAtlasInfo(atlas);
				iconString = CreateAtlasMarkup(atlas, 12, 12, 0, 0);
			end
		end
		
		local name = unlock.name;
		if (not unlock.validSpec) then
			name = name .. FORMAT_OFFSPEC_ICON;
		end
		
		tooltip:AddLine(iconString .. " " .. name, 1, 1, 1);
	end
end

ILW_DATA_MIXIN = {};

function ILW_DATA_MIXIN:Initialize()
	self.unlockedList = {};
	self.upcomming = {};
	self.upcommingLevel = 0;
	self.filteredList = {};
	self.filteredChart = {};
	self.unlockData = _addonData.unlockData;
	self.unlockTables = {};
	self.playerLevel = UnitLevel("player");
	self.fillerPool = CreateObjectPool(function() return {["type"] = UNLOCKTYPE_FILLER, ["noTooltip"] = true, ["id"] = 0} end);
	
	self.quickFilters = {
			[UNLOCKTYPE_DUNGEON] = function() return ILWAddon.db.global.trackers and ILWAddon.db.global.trackers["Dungeons"] end
			,[UNLOCKTYPE_RAID] = function() return ILWAddon.db.global.trackers and ILWAddon.db.global.trackers["Dungeons"] end
			,[UNLOCKTYPE_PVP] = function() return ILWAddon.db.global.trackers and ILWAddon.db.global.trackers["Battlegrounds"] end
			,[UNLOCKTYPE_RIDING] = function() return ILWAddon.db.global.trackers and ILWAddon.db.global.trackers["Riding"] end
			,[UNLOCKTYPE_SPELL] = function() return ILWAddon.db.global.trackers and ILWAddon.db.global.trackers["Spells"] end
			,[UNLOCKTYPE_TALENT] = function() return ILWAddon.db.global.trackers and ILWAddon.db.global.trackers["Talents"] end
			,[UNLOCKTYPE_UI] = function() return ILWAddon.db.global.trackers and ILWAddon.db.global.trackers["UI"] end
			,[UNLOCKTYPE_ZONES] = function() return ILWAddon.db.global.trackers and ILWAddon.db.global.trackers["Zones"] end
			,[UNLOCKTYPE_MISC] = function() return ILWAddon.db.global.trackers and ILWAddon.db.global.trackers["Misc"] end
			,["OffSpec"] = function(unlock) return (ILWAddon.db.global.trackers and ILWAddon.db.global.trackers["OffSpec"]) or unlock.validSpec end
		}
end

function ILW_DATA_MIXIN:PassesFilter(unlock)
	return self.quickFilters["OffSpec"](unlock) and (self.quickFilters[unlock.type] and self.quickFilters[unlock.type]() or unlock.subType == UNLOCK_SUBTYPE_UPCOMMING);
end

function ILW_DATA_MIXIN:LoadUnlockTable()
	wipe(self.unlockTables)
	
	local newTable = nil;
	for i = MINLEVEL, MAXLEVEL do
		if newTable == nil then
			newTable = {};
		end
		
		self:AddUnlocksOfLevel(i, newTable);
		
		if #newTable > 0 then
			self.unlockTables[i] = newTable;
			newTable = nil;
		end
	end
end

function ILW_DATA_MIXIN:UpdateFilteredList()
	wipe(self.filteredList);
	self.fillerPool:ReleaseAll();
	local levelCount = 0;
	local currentLevel;
	
	for k, unlock in ipairs(self.unlockedList) do
		if (self:PassesFilter(unlock)) then
			if (currentLevel and currentLevel ~= unlock.level) then
				if (levelCount % 2 ~= 0) then
					local filler = self.fillerPool:Acquire();
					filler.level = currentLevel;
					table.insert(self.filteredList, filler);
				end
				levelCount = 0;
			end
			table.insert(self.filteredList, unlock);
			currentLevel = unlock.level;
			levelCount = levelCount + 1;
		end
	end
end

function ILW_DATA_MIXIN:UpdateFilteredChart()
	wipe(self.filteredChart);
	for level, list in pairs(self.unlockTables) do
		if (level > MINLEVEL) then
			for k, unlock in ipairs(list) do
				if (self:PassesFilter(unlock)) then
					if not self.filteredChart[level] then
						self.filteredChart[level]  = {};
					end
					table.insert(self.filteredChart[level], unlock);
				end
			end
		end
	end
end

function ILW_DATA_MIXIN:ClearList()
	local list = self.unlockedList 
	for i = #list, 1, -1 do
		if list[i].subType ~= UNLOCK_SUBTYPE_UPCOMMING then
			table.remove(list, i);
		end
	end
	
	self.upcommingLevel = 0;
end

function ILW_DATA_MIXIN:LoadUnlocksForLevel(level, hideNew)
	self:ClearUpcommingFromList();
	local newUnlocks = self.unlockTables[level or self.playerLevel];
	local passedUnlocks = 0;
	if newUnlocks then
		for k, unlock in ipairs(newUnlocks) do
			unlock.read = hideNew;
			unlock.new = not hideNew;
			-- Skip if already on the list
			if (not self:AlreadyContainsUnlock(self.unlockedList, unlock)) then
				table.insert(self.unlockedList, unlock);
			end
			
			if (self:PassesFilter(unlock)) then
				passedUnlocks = passedUnlocks + 1;
			end
		end
		self:UpdateFilteredList();
	end
	
	return passedUnlocks;
end

function ILW_DATA_MIXIN:AddUnlocksOfLevel(level, list, ignoreSpec)
	list = list or self.unlockedList;
	local trackOptions = ILWAddon.db.global.trackers
	local startNr = #list;
	local unlocks;
	
	if (spec ~= nil) then
		ILWAddon.db.char.specLastLevel[spec] = level;
	end
	
	-- Check for skills
	unlocks = self.unlockData.skills[level]
	if (unlocks) then
		for k, unlock in ipairs(unlocks) do
			self:AddUnlock(list, UNLOCKTYPE_SPELL, unlock, level)
		end
	end
	
	-- Check for talents
	if (self.unlockData.talents[level]) then
		self:AddUnlock(list, UNLOCKTYPE_TALENT, nil, level);
	end

	-- Check for riding skills
	unlocks = self.unlockData.riding[level]
	if (unlocks) then
		for k, unlock in ipairs(unlocks) do
			self:AddUnlock(list, UNLOCKTYPE_RIDING, unlock, level)
		end
	end
	
	-- Check for UI
	unlocks = self.unlockData.UI[level]
	if (unlocks) then
		for k, unlock in ipairs(unlocks) do
			self:AddUnlock(list, UNLOCKTYPE_UI, unlock, level);
		end
	end
	
	-- Check for zones
	unlocks = self.unlockData.zones[level]
	if (unlocks) then
		for k, unlock in ipairs(unlocks) do
			self:AddUnlock(list, UNLOCKTYPE_ZONES, unlock, level);
		end
	end
	
	-- Check for dungeons
	unlocks = self.unlockData.instances[level]
	if (unlocks) then
		for k, unlock in ipairs(unlocks) do
			self:AddUnlock(list, (unlock.isRaid and UNLOCKTYPE_RAID or UNLOCKTYPE_DUNGEON), unlock, level);
		end
	end
	
	-- Check for PvP
	unlocks = self.unlockData.PVP[level]
	if (unlocks) then
		for k, unlock in ipairs(unlocks) do
			self:AddUnlock(list, UNLOCKTYPE_PVP, unlock, level);
		end
	end
	
	-- Check for misc
	unlocks = self.unlockData.misc[level]
	if (unlocks) then
		for k, unlock in ipairs(unlocks) do
			self:AddUnlock(list, UNLOCKTYPE_MISC, unlock, level);
		end
	end
	
	self:SortUnlockedList(list);
	
	return   #list > startNr;
end

function ILW_DATA_MIXIN:LoadUpcommingUnlocks(force)
	local level = self.playerLevel;
	if (not force and level == self.upcommingLevel) then return; end
	self.upcommingLevel = level;
	local numUnlocks = 0;
	local unlockTable = nil;
	wipe(self.upcomming)
	-- find the first next unlock
	while (level < MAXLEVEL and numUnlocks == 0) do
		level = level + 1;
		wipe(self.upcomming)
		unlockTable = self.unlockTables[level];
		if unlockTable then
			for k, unlock in ipairs(unlockTable) do
				if (self:PassesFilter(unlock)) then
					tinsert(self.upcomming, unlock)
				end
			end
			numUnlocks = #self.upcomming;
		end
	end
	
	if numUnlocks > 0 then
		local numToShow = numUnlocks > 2 and 1 or numUnlocks;
		
		local unlock
		for i=1, numToShow do
			unlock = self.upcomming[i];
			unlock.subType = UNLOCK_SUBTYPE_UPCOMMING;
			table.insert(self.unlockedList, unlock);
		end
		
		if (numToShow < numUnlocks) then
			local diff = numUnlocks - numToShow;
			UNLOCK_UPCOMMING_TRUNC.name = _L["FORMAT_UPCOMMING_SHORT"] :format(diff);
			UNLOCK_UPCOMMING_TRUNC.level = level;
			UNLOCK_UPCOMMING_TRUNC.subType = UNLOCK_SUBTYPE_UPCOMMING;
			wipe(UNLOCK_UPCOMMING_TRUNC.extra);
			for i = numToShow+1, numUnlocks do
				tinsert(UNLOCK_UPCOMMING_TRUNC.extra,  self.upcomming[i])
			end
			table.insert(self.unlockedList, UNLOCK_UPCOMMING_TRUNC);
		end
	end
	
	self:UpdateFilteredList();
	
	return self.upcomming and #self.upcomming or 0, level;
end

function ILW_DATA_MIXIN:AddUnlock(list, typeID, unlock, level)
	local id, icon, name, specs, subText, func, tooltip, atlas = 0;
	if unlock then
		id = unlock.id;
		icon = unlock.icon;
		name = unlock.name;
		subText = unlock.subText;
		tooltip = unlock.tooltip;
		atlas = unlock.atlas;
		func = unlock.func;
		extra = unlock.extra;
	end
	
	if (typeID == UNLOCKTYPE_SPELL or typeID == UNLOCKTYPE_RIDING) then
		local spellName, rank, spellIcon, castingTime, minRange, maxRange, spellID = GetSpellInfo(unlock.id);
		local rankText = GetSpellSubtext(unlock.id);
		local isPassive = IsPassiveSpell(unlock.id);
		subText = ""
		
		if typeID == UNLOCKTYPE_SPELL then
			local subTextInfo = {};
			if (isPassive) then
				tinsert(subTextInfo, SPELL_PASSIVE);
			else
				tinsert(subTextInfo, COMBATLOG_HIGHLIGHT_ABILITY);
			end
			if (rankText and rankText:match("%d")) then
				tinsert(subTextInfo, rankText);
			end
			if (unlock.subText) then
				tinsert(subTextInfo, unlock.subText);
			end
			subText = table.concat(subTextInfo, ", ");
		elseif typeID == UNLOCKTYPE_RIDING then
			subText = _L["REQUIRES TRAINING"];
		end
		name = spellName;
		icon = spellIcon;
		specs = unlock.specs;
		
	elseif (typeID == UNLOCKTYPE_TALENT) then
		icon = ICON_TALENT;
		name = _L["TALENT_POINT"];
		subText = TALENT;
	
	elseif (typeID == UNLOCKTYPE_UI) then
		icon = ICON_UI;
	elseif (typeID == UNLOCKTYPE_ZONES) then
		name =  C_Map.GetMapInfo(unlock.id).name;
		subText = ZONE .. ", " .. C_Map.GetMapInfo(unlock.continentID).name;
	end
	
	if((not unlock or not unlock.noTooltip) and name == nil) then
		print("[ILW] Nil name for", unlock.id, level, typeID);
		return;
	end
	local unlock = {["type"] = typeID, ["level"] = level, ["id"] = id, ["icon"] = icon, ["name"] = name ,["specs"] = specs
			, ["func"] = func, ["subText"] = subText, ["tooltip"] = tooltip, ["atlas"] = atlas, ["extra"] = extra};
	unlock.validSpec = IsValidSpec(unlock);
	table.insert(list, unlock);
	ILW_UnlockContainer:UpdateUnlockDisplay();
end

function ILW_DATA_MIXIN:AlreadyContainsUnlock(list, toCheck)
	for key, unlock in ipairs(list) do
		if (unlock.id == toCheck.id and unlock.type == toCheck.type) then
			return true;
		end
	end
	return false;
end

function ILW_DATA_MIXIN:ClearUpcommingFromList()
	for i = #self.unlockedList, 1, -1 do
		if (self.unlockedList[i].subType == UNLOCK_SUBTYPE_UPCOMMING) then
			self.unlockedList[i].subType = nil;
			table.remove(self.unlockedList, i);
		end
	end
end

function ILW_DATA_MIXIN:SortUnlockedList(list)
	table.sort(list or self.unlockedList, function(a, b)
			if (a.level == b.level) then
				if (a.type == b.type) then
					if(a.validSpec ~= b.validSpec) then
						return a.validSpec and not b.validSpec;
					end
					return a.id > b.id;
				end
				return a.type < b.type;
			end
			return a.level > b.level;
		end
	);
	
	self:UpdateFilteredList();
end

function ILW_DATA_MIXIN:ValidateUnlockSpec()
	for k, unlock in ipairs(self.unlockedList) do
		unlock.validSpec = IsValidSpec(unlock);
	end
end

function ILW_DATA_MIXIN:GetLevelPage(level)
	for index, info in ipairs(self.filteredList) do
		if (info.level == level) then
			return ceil(index/UNLOCKS_PER_PAGE);
		end
	end
	return 1;
end

function ILW_DATA_MIXIN:MarkLevelAsNew(level)
	for index, info in ipairs(self.unlockedList) do
		info.new = info.level == level;
	end
end


ILW_UNLOCK_MIXIN = {};

-- When the player clicks on the icon of an unlock
function ILW_UNLOCK_MIXIN:OnClick(button)
	-- prevent 'checked' visualf
	self:SetChecked(false);

	if (self.data.subType == UNLOCK_SUBTYPE_UPCOMMING) then
		return;
	end
	
	-- Don't have to open anything if it's right mouse
	if (button == "RightButton") then
		
		ILW_UnlockContainer:UpdateUnlockDisplay();
		return;
	end
	-- Open spellbook if spell
	if (self.unlockType == UNLOCKTYPE_SPELL) then
		if (not IsPassiveSpell(self.unlockId)) then
			PickupSpell(self.unlockId);
		end
	-- Open talents
	elseif (self.unlockType == UNLOCKTYPE_TALENT) then
		if (PlayerTalentFrame == nil) then
			LoadAddOn("Blizzard_TalentUI");
		end
		ShowUIPanel(PlayerTalentFrame);
	-- Open encounter journal if instance
	elseif (self.unlockType == UNLOCKTYPE_DUNGEON or self.unlockType == UNLOCKTYPE_RAID) then
		if (EncounterJournal == nil) then
			LoadAddOn("Blizzard_EncounterJournal");
		end
		ShowUIPanel(EncounterJournal);
		if (self.unlockId ~= nil) then
			
			-- id > 0 = specific instance so open journal to that instance
			-- otherwise see _aVar and open to instance list
			if (self.unlockId > 0) then
				-- actual instance so open journal to specific instance
				EncounterJournal_DisplayInstance(self.unlockId);
			elseif (self.data.extra) then
				EncounterJournal_TierDropDown_Select(_, self.data.extra.tier)
				EJ_ContentTab_Select(self.data.extra.tab);
			end
		end
	elseif (self.unlockType == UNLOCKTYPE_PVP) then
		ShowUIPanel(PVEFrame);
		PVEFrame_ShowFrame("PVPUIFrame", "HonorFrame");
	elseif (self.unlockType == UNLOCKTYPE_UI) then
		if (self.data.func) then 
			self.data.func()
		end
	elseif (self.unlockType == UNLOCKTYPE_ZONES) then
		ShowUIPanel(WorldMapFrame);
		WorldMapFrame:SetMapID(self.data.id);
	elseif self.data.func then
		self.data.func();
	end
end

function ILW_UNLOCK_MIXIN:OnDragStart()
	if (self.data.subType == UNLOCK_SUBTYPE_UPCOMMING) then
		return;
	end
	if (self.unlockType == UNLOCKTYPE_SPELL) then
		if (not IsPassiveSpell(self.unlockId)) then
			PickupSpell(self.unlockId);
		end
	end
end

function ILW_UNLOCK_MIXIN:OnEnter(motion)
	local info = self.data;
	if (info.typeID == UNLOCKTYPE_FILLER) then return end;

	-- show tooltip depending on type
	GameTooltip:SetOwner(self, "ANCHOR_RIGHT");

	if (self.unlockType == UNLOCKTYPE_SPELL) then
		GameTooltip:SetSpellByID(self.unlockId);
	elseif (self.unlockType == UNLOCKTYPE_UPCOMMING) then
		GameTooltip:SetText(self.data.name, 1, 1, 1);
		AddUnlocksToTooltip(GameTooltip, info.extra)
	else
		GameTooltip:SetText(self.data.name, 1, 1, 1);
		local tooltip = self.data.tooltip;
		if (self.unlockType == UNLOCKTYPE_RIDING) then
			tooltip =_L["TOOLTIP_RIDING"];
		elseif (self.unlockType == UNLOCKTYPE_TALENT) then
			tooltip = _L["TOOLTIP_TALENT"];
		elseif (self.unlockType == UNLOCKTYPE_DUNGEON or self.unlockType == UNLOCKTYPE_RAID) then
			tooltip = _L["TOOLTIP_INSTANCE"];
		elseif (self.unlockType == UNLOCKTYPE_PVP) then
			tooltip = _L["TOOLTIP_PVP"];
		elseif (self.unlockType == UNLOCKTYPE_UI) then
			tooltip = _L["TOOLTIP_UI"];
		elseif (self.unlockType == UNLOCKTYPE_ZONES) then
			tooltip = _L["TOOLTIP_ZONE"];
		end
		GameTooltip:AddLine(tooltip, 1, .82, 0, true);
	end

	if (info.subType == UNLOCK_SUBTYPE_UPCOMMING) then
		local text = _L["TOOLTIP_UPCOMMING_NEXT"];
		local diff = info.level - UnitLevel("player");
		if (diff > 1) then
			text = _L["TOOLTIP_UPCOMMING"]:format(diff);
		end
		GameTooltip:AddLine(text, .75, .75, .75, true);
	end
	if (not info.validSpec) then
		GameTooltip:AddLine(_L["NOT_CURRENT_SPEC"], 0.85, 0, 0, true);
	end
	
	GameTooltip:Show();
	
	-- hide the "New!" tag and mark as read to prevent animation
	self.NewText:SetText("");
	self.NewTextBG:Hide();
	info.read = true;
	
end

function ILW_UNLOCK_MIXIN:Reset()
	local name = self:GetName();
	local iconTexture = _G[name.."IconTexture"];
	local slotFrame = _G[name.."SlotFrame"];
	self.animation:Stop();
	iconTexture:Hide();
	iconTexture:SetTexture("");
	slotFrame:SetAlpha(0);
	self.Blocker:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT");
	self.TypeIcon:SetAlpha(0);
	self.SpellName:SetText("");
	self.SpellName:SetHeight(self.SpellName:GetStringHeight());
	self.unlockType = nil;
	self.SpellSubName:SetText("");
	self.unlockId = nil;
	self.read = false;
	self.new = true;
	self.NewText:SetText("");
	self.NewTextBG:Hide();
	self.Level:Hide();
	self.DigitBg:Hide();
	self:SetAlpha(0.5);
	self.InvalidOverlay:SetAlpha(0);
end

function ILW_UNLOCK_MIXIN:Update(unlock, levelCount)
	self.data = unlock;
	self.unlockType = unlock.type;
	
	if (unlock.type == UNLOCKTYPE_FILLER) then
		return;
	end

	self:SetAlpha(1);
	self.Blocker:SetPoint("BOTTOMRIGHT", self, "TOPLEFT");
	
	local name = self:GetName();
	local iconTexture = _G[name.."IconTexture"];
	local slotFrame = _G[name.."SlotFrame"];
	local parent = self:GetParent();
	iconTexture:Show();
	iconTexture:SetTexture(unlock.icon);
	slotFrame:SetAlpha(1);
	self.unlockId = unlock.id;
	
	-- check if unlock has been read before.
	self.read = unlock.read;

	self.SpellName:SetText(unlock.name);
	self.SpellName:Show();
	self.SpellName:SetHeight(self.SpellName:GetStringHeight());
	
	local atlas = unlock.atlas or _typeIcons[self.unlockType];
	if atlas then
		self.TypeIcon:SetAtlas(atlas)
	end
	
	local subText = "";

	if (unlock.subText ~= nil) then
		subText = subText .. unlock.subText;
	end
	
	if (unlock.specs) then
		local specId = GetPlayerSpec();
		local hasCurrentSpec = unlock.validSpec;
		if(specId and hasCurrentSpec) then
			local specName = select(2, GetSpecializationInfoByID(specId));
			subText = subText ..", ".. specName;
		end
		
		for k, spec in ipairs(unlock.specs) do
			if (spec ~= specId) then
				local specName = select(2, GetSpecializationInfoByID(spec));
				if (not hasCurrentSpec) then
					specName = FORMAT_OFFSPEC:format(specName);
				end
				subText = subText ..", ".. specName;
			end
		end
		
		if (not hasCurrentSpec) then
			self.InvalidOverlay:SetAlpha(1);
		end
	end
	
	self.TypeIcon:SetAlpha(subText == "" and 0 or 1);
	self.SpellSubName:SetText(subText);
	self.SpellSubName:Show();
	self.SpellSubName:SetHeight(self.SpellSubName:GetStringHeight());
	
	-- Only show level for the first one
	if (levelCount == 1) then
		self.DigitBg:Show();
		self.Level:Show();
		self.Level:SetText(unlock.level);
		
		local id = self:GetID();
		if (id > 1) then
			local index = floor(id/2);
			parent.dividers[index]:Show();
		end
	end
	
	if (unlock.subType == UNLOCK_SUBTYPE_UPCOMMING) then
		iconTexture:SetDesaturated(true);
		self.SlotFrame:SetDesaturated(true);
		self.TypeIcon:SetDesaturated(true);
		self.Level:SetTextColor(0.7, 0.7, 0.7, 1);
		self.SpellName:SetTextColor(0.7, 0.7, 0.7, 1);
		self.TextBackground:SetDesaturated(true);
		
	else
		iconTexture:SetDesaturated(false);
		self.SlotFrame:SetDesaturated(false);
		self.TypeIcon:SetDesaturated(false);
		self.Level:SetTextColor(1, 0.82, 0, 1);
		self.SpellName:SetTextColor(1, 0.82, 0, 1);
		self.TextBackground:SetDesaturated(false);
		
		unlock.new = (unlock.new == nil and true or unlock.new );
		if (not self.read) then
		
			self.NewText:SetText(_L["NEW"]);
			if (self.NewText:GetWidth() > self:GetWidth()) then
				self.NewText:SetTextHeight(13);
			end
			self.NewTextBG:Show();
		end
		
		if unlock.new then
			self.animation:Play();
			unlock.new = false;
		end

	end
end




ILW_CORE_MIXIN = {}

function ILW_CORE_MIXIN:OnShow()
	HideUIPanel(ILW_AlertPopup);
	self:UpdateUnlockDisplay();
	PlaySound(SOUNDKIT.IG_SPELLBOOK_OPEN);
	
	-- Update chart
	ILW_Chart:SetToLevel();
end

function ILW_CORE_MIXIN:OnHide()
	HideUIPanel(ILW_HelpFrame);
	ILW_ShowHelpUnlocks(false);
	PlaySound(SOUNDKIT.IG_SPELLBOOK_CLOSE);
end

function ILW_CORE_MIXIN:OnLoad()

	self.dataProvider = CreateFromMixins(ILW_DATA_MIXIN);
	
	self:RegisterForDrag("LeftButton");
	self:SetScript("OnDragStart", self.StartMoving );
	self:SetScript("OnDragStop", self.StopMovingOrSizing);
	-- allows the player to close the frame using Esc like regular blizzard windows
	table.insert(UISpecialFrames, "ILW_UnlockContainer");
	UIPanelWindows["ILW_UnlockContainer"] = { area = "left", pushable = 4 };
	
	self.Navigation.CurrentPage = 1;
	
	for i=1, UNLOCKS_PER_PAGE do
		local button = _G["ILW_UnlockContainerUnlock"..i];
		button:RegisterForDrag("LeftButton");
		button:RegisterForClicks("LeftButtonUp", "RightButtonUp");
		button.read = false;
	end
	
	SetPortraitToTexture(ILW_UnlockContainerPortrait, "Interface\\ICONS\\Spell_Holy_SurgeOfLight");
	ILW_UnlockContainerTitleText:SetText("I Learned What?")
	
	ButtonFrameTemplate_HideButtonBar(self);
	ButtonFrameTemplate_HideAttic(self);
end

function ILW_CORE_MIXIN:OnMouseWheel(delta)
	if delta == 1 then
		self.Navigation:PrevPageButton_OnClick();
	else
		self.Navigation:NextPageButton_OnClick();
	end
end

function ILW_CORE_MIXIN:Open()
	ShowUIPanel(self);
end

function ILW_CORE_MIXIN:LevelUp(level, skipChartUpdate, silent)
	local numUnlocks = self.dataProvider:LoadUnlocksForLevel(level);
	self.dataProvider:LoadUpcommingUnlocks();
	self.dataProvider:SortUnlockedList();
	if (ILW_UnlockContainer:IsShown()) then
		self:UpdateUnlockDisplay();
		if (not skipChartUpdate) then
			ILW_Chart:SetToLevel(level);
		end
	else 
		if (not silent and numUnlocks > 0) then
			ILW_AlertPopup:Present(numUnlocks);
		end
	end
end

function ILW_CORE_MIXIN:ResetButtons()
	for i=1, UNLOCKS_PER_PAGE do
		local button = _G["ILW_UnlockContainerUnlock"..i];
		button:Reset();
	end
	
	for k, divider in ipairs(self.dividers) do
		divider:Hide();
	end
	
	self.Navigation:Update();
end

function ILW_CORE_MIXIN:UpdateUnlockDisplay()
	if (not self:IsShown()) then return; end
	
	self:ResetButtons();
	self.dataProvider:UpdateFilteredList();
	local unlockedList = self.dataProvider.filteredList;
	local prevDisplayLevel = 0;
	local count = 1;
	local unlock = nil;
	local pageNr = self.Navigation.CurrentPage;
	local start = (pageNr-1) * UNLOCKS_PER_PAGE;
	local nrToShow = (#unlockedList-start) > UNLOCKS_PER_PAGE and UNLOCKS_PER_PAGE or #unlockedList-start;
	local levelCount = 1;
	for i = start + 1, (start + nrToShow) do
		unlock = unlockedList[i];
		if (prevDisplayLevel ~= unlock.level) then
			levelCount = 1;
		end
		
		local button = _G["ILW_UnlockContainerUnlock"..count];
		button:Update(unlock, levelCount);
		levelCount = levelCount + 1;
			
		prevDisplayLevel = unlock.level;
			
		count = count +1;
	end

	self.Navigation:Update();
end

function ILW_CORE_MIXIN:SpecChanged()
	self.dataProvider:ValidateUnlockSpec();
	self.dataProvider:SortUnlockedList();
	self.dataProvider:UpdateFilteredChart();
	self.Chart:Update();
	
	self.dataProvider:UpdateFilteredList();
	
	self:UpdateUnlockDisplay();
end

function ILW_CORE_MIXIN:UpdateChartTooltip()
	local hightledBlock = self.Chart.hightledBlock;
	if (hightledBlock) == nil then return; end
	GameTooltip:SetOwner(hightledBlock, "ANCHOR_RIGHT");
	GameTooltip:SetText(UNIT_LEVEL_TEMPLATE:format(hightledBlock.blockLevel));
	
	if (not hightledBlock.unlocks) then return end;

	AddUnlocksToTooltip(GameTooltip, hightledBlock.unlocks)
	
	GameTooltip:Show();
end

function ILW_CORE_MIXIN:TutorialButton_OnClick()
	if InCombatLockdown() then return; end
	if ILW_HelpFrame:IsShown() then
		_help:HideTutorial();
		ILW_ShowHelpUnlocks(false);
	else
		_help:ShowTutorial();
		ILW_ShowHelpUnlocks(true);
	end
end

function ILW_CORE_MIXIN:FiltersUpdated()
	local data = self.dataProvider;
	data:UpdateFilteredChart();
	data:ClearUpcommingFromList();
	data:LoadUpcommingUnlocks(true);
	data:SortUnlockedList();
	self:UpdateUnlockDisplay();
	ILW_Chart:Update();
end



ILW_NAVIGATION_MIXIN = {};

function ILW_NAVIGATION_MIXIN:PrevPageButton_OnClick()
	if (self.CurrentPage == 1) then return; end
	PlaySound(SOUNDKIT.IG_ABILITY_PAGE_TURN);
	self.CurrentPage = self.CurrentPage - 1;
	self.Text:SetText(PAGE_NUMBER:format(self.CurrentPage));
	ILW_UnlockContainer:UpdateUnlockDisplay();
end

function ILW_NAVIGATION_MIXIN:NextPageButton_OnClick()
	if (self.CurrentPage >= ceil(#ILW_UnlockContainer.dataProvider.filteredList/UNLOCKS_PER_PAGE)) then return; end
	PlaySound(SOUNDKIT.IG_ABILITY_PAGE_TURN);
	self.CurrentPage = self.CurrentPage + 1;
	self.Text:SetText(PAGE_NUMBER:format(self.CurrentPage));
	ILW_UnlockContainer:UpdateUnlockDisplay();
end

function ILW_NAVIGATION_MIXIN:Update()
	local totalPages = ceil(#ILW_UnlockContainer.dataProvider.filteredList/UNLOCKS_PER_PAGE);

	if (totalPages > 0 and self.CurrentPage > totalPages) then
		self.CurrentPage = totalPages;
	elseif (totalPages == 0) then
		self.CurrentPage = 1;
	end

	self.Text:SetText(PAGE_NUMBER:format(self.CurrentPage));
	
	self.Prev:Enable();
	self.Next:Enable();

	-- disable prev on first page;
	if (self.CurrentPage == 1) then
		self.Prev:Disable();
	end

	-- disable next page if on last page
	if (totalPages == 0 or self.CurrentPage == totalPages) then
		self.Next:Disable();
	end
end

function ILW_NAVIGATION_MIXIN:OpenToPage(page)
	self.CurrentPage = page;
	PlaySound(SOUNDKIT.IG_ABILITY_PAGE_TURN);
	self:Update();
end



function ILWAddon:InitFilter(dd, level)
	local info = ADD:CreateInfo();
	info.keepShownOnClick = true;	

	if level == 1 then
		info.text = _L["OPTION_MINIMAP"];
		info.func = function(_, _, _, value)
						ILWAddon.db.global.minimap.hide = not value;
						if (ILWAddon.db.global.minimap.hide) then
							_icon:Hide(_addonName);
						else
							_icon:Show(_addonName);
						end
					end 
		info.checked = not ILWAddon.db.global.minimap.hide;
		info.isNotRadio = true;
		ADD:AddButton(info, level)

		info.text = _L["OPTION_POPUP"];
		info.func = function(_, _, _, value)
						ILWAddon.db.global.showPopup = value;
					end 
		info.checked = ILWAddon.db.global.showPopup;
		info.isNotRadio = true;
		ADD:AddButton(info, level)
	
		info.checked = 	nil;
		info.isNotRadio = nil;
		info.func =  nil;
		info.hasArrow = true;
		info.notCheckable = true;
		
		info.text = _L["OPTION_TRACKERS"];
		info.value = 1;
		ADD:AddButton(info, level)
	else --if level == 2 then
		
		info.hasArrow = false;
		info.isNotRadio = true;
		info.notCheckable = true;
			
		info.text = CHECK_ALL
		info.func = function()
						ILWAddon:SetAllSourcesTo(true);
						ADD:Refresh(dd, 1, 2);
						ILW_UnlockContainer:FiltersUpdated();
					end
		ADD:AddButton(info, level)
		
		info.text = UNCHECK_ALL
		info.func = function()
						ILWAddon:SetAllSourcesTo(false);
						ADD:Refresh(dd, 1, 2);
						ILW_UnlockContainer:FiltersUpdated();
					end
		ADD:AddButton(info, level)

		
		info.notCheckable = false;
		for k, v in pairs(_trackers) do
			info.text = _trackerLabels[k];
			info.func = function(_, _, _, value)
								ILWAddon.db.global.trackers[k] = value;
								ILW_UnlockContainer:FiltersUpdated();
							end
			info.checked = function() return ILWAddon.db.global.trackers[k] end;
			ADD:AddButton(info, level);			
		end
	end
end

function ILWAddon:SetAllSourcesTo(enable)
	for k, v in pairs(ILWAddon.db.global.trackers) do
		ILWAddon.db.global.trackers[k] = enable;
	end
end





ILW_POPUP_MIXIN = {};

function ILW_POPUP_MIXIN:OnShow()
	self.animLightUp:Play();
end

function ILW_POPUP_MIXIN:OnClick(button)
	if(button == "LeftButton") then
		ILW_UnlockContainer:Open();
	end
	HideUIPanel(self)
end

function ILW_POPUP_MIXIN:Present(num)
	if (num and (num > self.num)) then
		self.num = num;
	end
	if (ILW_UnlockContainer:IsShown() or self.num == 0 or not ILWAddon.db.global.showPopup ) then return; end
	if (InCombatLockdown()) then
		self.shownDuringCombat = true;
		return;
	end
	self.text:SetText(_L["POPUP_FORMAT"]:format(self.num));
	ShowUIPanel(self);
	self.num  = 0;
end





local NUM_CHART_BLOCKS = 23;

ILW_CHART_MIXIN = {}

function ILW_CHART_MIXIN:OnLoad()
	self:SetFrameLevel(ILW_UnlockContainer:GetFrameLevel()+3);
	
	local lastBlock = nil;
	for i=1, NUM_CHART_BLOCKS do
		local block = self:GetOrCreateBlock(i);
		if lastBlock then
			block:SetPoint("LEFT", lastBlock, "RIGHT", 0, 0);
		else
			block:SetPoint("TOPLEFT", self, "TOPLEFT", 3, -5);
		end
		lastBlock = block;
	end
	
	self.offset = 0;
	self.playerLevel = UnitLevel("player");
	self.playerSpec = GetPlayerSpec();
	
	ILW_Chart.Slider:SetObeyStepOnDrag(true);
end

function ILW_CHART_MIXIN:OnMouseWheel(delta)
	self.offset = self.offset + delta * -1;
	self.offset = max(self.offset, MINLEVEL-1);
	self.offset = min(self.offset, MAXLEVEL - NUM_CHART_BLOCKS);
	
	self:Update();
	ILW_UnlockContainer:UpdateChartTooltip();
end

function ILW_CHART_MIXIN:GetOrCreateBlock(index)
	local block = self.BlockArray and self.BlockArray[index];
	if block then
		block:Show();
		return block;
	end

	block = CreateFrame("FRAME", nil, self, "ILW_ChartBlock_Template");
	block.unlocks = {};
	block:SetWidth((622-6)/NUM_CHART_BLOCKS)

	return block;
end

function ILW_CHART_MIXIN:Update(fromSlider)
	if (not self.BlockArray) then return; end
	self.offset = max(self.offset, MINLEVEL-1);
	self.offset = min(self.offset, MAXLEVEL - NUM_CHART_BLOCKS);
	if (not fromSlider) then
		self.Slider:SetValue(self.offset);
		local minValue, maxValue = self.Slider:GetMinMaxValues();
	end
	
	for i=1, NUM_CHART_BLOCKS do
		local block = self.BlockArray[i];
		block.blockLevel = i + self.offset;
		block:Update();
	end
	
	self.Slider:UpdateProgress();
end

function ILW_CHART_MIXIN:SetToLevel(level)
	self.playerLevel = ILW_UnlockContainer.dataProvider.playerLevel;
	local minValue, maxValue = self.Slider:GetMinMaxValues();
	
	local value = self.playerLevel - 5;
	value = max(value, minValue);
	value = min(value, maxValue);
	self.offset = value;
	self:Update();
end




ILW_CHART_BLOCK_MIXIN = {}

function ILW_CHART_BLOCK_MIXIN:Update()
	local lightUp = self.blockLevel <= self:GetParent().playerLevel;
	self.Progress:SetAlpha(lightUp and 1 or 0);
	
	-- levels
	self.Level:SetText("");
	self.LevelFlag:SetAlpha(0);
	if (self.blockLevel % 5 == 0) then
		self.Level:SetText(self.blockLevel);
		self.LevelFlag:SetAlpha(1);
		self.Level:SetFontObject(lightUp and GameFontNormalSmall or ILW_GameFontLightGraySmall);
		if (lightUp) then
			self.LevelFlag:SetVertexColor(1, 0.82, 0);
		else
			self.LevelFlag:SetVertexColor(1, 1, 1);
		end
	end
	
	-- unlocks
	self.unlocks = ILW_UnlockContainer.dataProvider.filteredChart[self.blockLevel];
	
	self.Number:SetText("");
	self.NumberBG:SetAlpha(0);
	self.NumberBG2:SetAlpha(0);
	if (self.unlocks) then
		self.Number:SetText(#self.unlocks);
		self.NumberBG:SetAlpha(0.75);
		
		self.NumberBG2:SetDesaturated(not lightUp);
		self.NumberBG2:SetAlpha(1);
		self.Number:SetFontObject(lightUp and GameFontNormalSmall or ILW_GameFontLightGraySmall);
	end
	
end

function ILW_CHART_BLOCK_MIXIN:OnEnter()
	
	self.HighlightTop:SetAlpha(1);
	self.HighlightBottom:SetAlpha(1);
	
	self:GetParent().hightledBlock = self;
	
	ILW_UnlockContainer:UpdateChartTooltip();
end

function ILW_CHART_BLOCK_MIXIN:OnLeave()
	self:GetParent().hightledBlock = nil;
	GameTooltip:Hide();
	self.HighlightTop:SetAlpha(0);
	self.HighlightBottom:SetAlpha(0);
end

function ILW_CHART_BLOCK_MIXIN:OnMouseUp(button)
	if (button ~= "LeftButton" or  self.blockLevel > ILW_UnlockContainer.dataProvider.playerLevel or not self.unlocks or #self.unlocks == 0) then
		return;
	end
	local page = ILW_UnlockContainer.dataProvider:GetLevelPage(self.blockLevel)
	ILW_UnlockContainer.dataProvider:MarkLevelAsNew(self.blockLevel);
	ILW_UnlockContainer.Navigation:OpenToPage(page);
	ILW_UnlockContainer:UpdateUnlockDisplay();
end



ILW_CHARTSLIDER_MIXIN = {}

function ILW_CHARTSLIDER_MIXIN:OnLoad()
	self:SetOrientation("HORIZONTAL")
	
	local maxValue = max(NUM_CHART_BLOCKS, MAXLEVEL - NUM_CHART_BLOCKS)
	local minValue = min(MINLEVEL-1, maxValue);
	self:SetMinMaxValues(minValue, maxValue);
	self:SetValueStep(1);
	self:SetValue(UnitLevel("player")-5);
	self.ThumbOverlay:SetPoint("CENTER", self.ThumbTexture, "CENTER", -1, 0);
	
end

function ILW_CHARTSLIDER_MIXIN:OnValueChanged(value, playerSet)
	local parent = self:GetParent();
	value = floor(value);
	if (not playerSet or value == parent.offset) then return; end

	self.ThumbOverlay:SetPoint("CENTER", self.ThumbTexture, "CENTER", -1, 0);
	
	ILW_Chart.offset = value;
	ILW_Chart:Update(true);
end

function ILW_CHARTSLIDER_MIXIN:UpdateProgress(level)
	local width = self:GetWidth()-1;
	local progress = ((level or UnitLevel("player"))- MINLEVEL+1) / (MAXLEVEL - MINLEVEL+1);
	self.Progress:SetWidth(progress * width)
end





function ILWAddon:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("ILWDB", _defaults, true);
	_icon:Register(_addonName, ILWBroker, self.db.global.minimap);
	
	ILW_UnlockContainer.dataProvider:Initialize();
	if (not self.db.global.version) then
		wipe(ILW_UnlockContainer.dataProvider.unlockedList)
	end
	self.db.global.version = _version;
end

function ILWAddon:OnEnable()
	local playerLevel =  UnitLevel("player");
	local spec = GetPlayerSpec();
	if (spec ~= nil) then
		self.db.char.specLastLevel[spec] = playerLevel;
	end
	
	local frame = ADD:CreateMenuTemplate("ILW_UnlockContainerOptionsDropDown", ILW_UnlockContainer);
	frame.noResize = true;
	
	ADD:Initialize(frame, function(self, level) ILWAddon:InitFilter(self, level) end, "MENU");
	
	-- Get official names for dungeons, battlegounds and worldPVP zones
	-- This has to be done here as some info (mainly BGs) is not available before the addon is done loading.
	-- Get names for official instances
	for level, instances in pairs(_addonData.unlockData.instances) do
		for k, instance in ipairs(instances) do
			if (instance.subText == ENCOUNTER_JOURNAL_INSTANCE) then
				instance.name = EJ_GetInstanceInfo(instance.id);
			end
		end
	end
	
	-- Official names
	for level, unlocks in pairs(_addonData.unlockData.PVP) do
		for k, unlock in ipairs(unlocks) do
			if (unlock.subText == BATTLEGROUND) then
				unlock.name = GetBattlegroundInfo(unlock.id);
			elseif (unlock.subText == _L["WORLD_PVP"]) then
				unlock.name = select(2, GetWorldPVPAreaInfo(unlock.id));
			end
		end
	end
	
	ILW_UnlockContainer.dataProvider:LoadUnlockTable();
	for i = MINLEVEL+1, playerLevel do 
		ILW_UnlockContainer.dataProvider:LoadUnlocksForLevel(i, true);
	end
	ILW_UnlockContainer.dataProvider:LoadUpcommingUnlocks();
	ILW_UnlockContainer.dataProvider:SortUnlockedList();
	ILW_UnlockContainer.dataProvider:UpdateFilteredChart();
	
	-- Opening the frame once allows it to be used during combat. for some reason
	ShowUIPanel(ILW_UnlockContainer);
	HideUIPanel(ILW_UnlockContainer);
end

local L_ILWhat_LoadFrame = CreateFrame("FRAME", "ILWhat_LoadFrame"); 
ILWhat_LoadFrame:RegisterEvent("PLAYER_LEVEL_UP");
ILWhat_LoadFrame:RegisterEvent("PLAYER_REGEN_ENABLED");
ILWhat_LoadFrame:RegisterEvent("PLAYER_REGEN_DISABLED");
ILWhat_LoadFrame:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED");
ILWhat_LoadFrame:RegisterEvent("ADDON_LOADED");
ILWhat_LoadFrame:RegisterEvent("PLAYER_LOGOUT");
ILWhat_LoadFrame:SetScript("OnEvent", function(self, event, ...) self[event](self, ...) end)

function ILWhat_LoadFrame:PLAYER_LEVEL_UP(level, hp, mp, talentPoints, strength, agility, stamina, intellect, spirit)
	ILW_UnlockContainer.dataProvider.playerLevel = level;

	ILW_UnlockContainer:LevelUp(level);
end

function ILWhat_LoadFrame:PLAYER_REGEN_DISABLED()
	ILW_AlertPopup:Hide();
	HideUIPanel(ILW_UnlockContainer);
end

function ILWhat_LoadFrame:PLAYER_REGEN_ENABLED()
	if ILW_AlertPopup.shownDuringCombat then
		ILW_AlertPopup:Present();
		ILW_AlertPopup.shownDuringCombat = false;
	end
end

function ILWhat_LoadFrame:PLAYER_SPECIALIZATION_CHANGED(player)
	if(player == "player") then
		ILW_UnlockContainer:SpecChanged();
	end
end

function ILWhat_LoadFrame:PLAYER_LOGOUT(loadedAddon)
	ILW_ShowHelpUnlocks(false);
end

function ILWhat_LoadFrame:ADDON_LOADED(loadedAddon)
	if (loadedAddon ~= _addonName) then return; end
	
	ILWhat_LoadFrame:UnregisterEvent("ADDON_LOADED")
	
	_help:Initialize(ILW_UnlockContainer, _helpPlate);

end

----------------------------------------
-- Slash Commands
----------------------------------------

SLASH_ILEARNEDWHAT1 = '/ilwhat';
SLASH_ILEARNEDWHAT2 = '/ilearnedwhat';
SLASH_ILEARNEDWHAT3 = '/ilw';
local function slashcmd(msg, editbox)
	-- This parses the character's spellbook
	-- local class, classFilename, classID = UnitClass("player");
	-- class = strlower(class);
	-- local s,id=MINLEVEL+1,1;
	-- local t={}; 
	-- for i=s,120 do
		-- t[i] = {}
	-- end
	
	-- local tab = 2
	-- local _, _, offset, numEntries, _, _, _,spec =  GetSpellTabInfo(tab)
	-- while spec do
		-- local start = offset+1
		-- for i=start, start+numEntries do
			-- local name,_,id = GetSpellBookItemName(i, "spell") 
			-- local level = GetSpellLevelLearned(id); 
			
			-- if id and level > MINLEVEL then
				-- if (not t[level][id]) then
					-- t[level][id] = {["name"] = name, ["specs"] = {spec}};
				-- else
					-- tinsert(t[level][id].specs, spec);
				-- end
			-- end
		-- end
		-- tab = tab + 1;
		-- _, _, offset, numEntries, _, _, _,spec  =GetSpellTabInfo(tab)
	-- end
	-- local numSpecs = GetNumSpecializationsForClassID(classID);
	-- local text = "";
	-- for k,v in pairs(t) do 
		-- for k2,info in pairs(v) do 
			-- local subtext = GetSpellSubtext(k2) or "";
			-- local o = "AddUnlock("..(msg=="" and class or msg).."Skills, "..k..", {[\"id\"] = "..k2;
			-- if (#info.specs < numSpecs ) then
				-- o = o..", [\"specs\"] = {";
				-- o = o..info.specs[1];
				-- for i=2, #info.specs do
					-- o = o..", "..info.specs[i];
				-- end
				-- o = o.."}";
			-- end
			-- o = o.."}); -- "..info.name .. " " ..(subtext:match("%d") or "")
			-- text = text ..o .."\n"
		-- end
	-- end
	-- ILW_Output:SetText(text)
	-- ILW_Output:Show();
	-- if true then return; end
	-- end of parsing
	

	if (ILW_UnlockContainer:IsShown()) then
		HideUIPanel(ILW_UnlockContainer);
	else
		ILW_UnlockContainer:Open();
	end
end
SlashCmdList["ILEARNEDWHAT"] = slashcmd